"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.BaseDocument = exports.SortDirection = void 0;
var SortDirection;
(function (SortDirection) {
    SortDirection["ASC"] = "asc";
    SortDirection["DESC"] = "desc";
})(SortDirection || (exports.SortDirection = SortDirection = {}));
class BaseDocument {
    constructor(data) {
        Object.keys(this.constructor.prototype.__fields).forEach((i) => {
            const field = this.constructor.prototype.__fields[i];
            if (field.defaultValue) {
                this[i] = field.defaultValue;
            }
        });
        if (data) {
            this.apply(data);
        }
    }
    async get() {
        try {
            const response = await this.constructor.prototype.client.get({
                id: this.id,
                index: this.constructor.prototype.__indexName,
                type: this.constructor.prototype.__type,
                realtime: true,
                refresh: true,
            });
            this.apply(response.body._source);
            return this;
        }
        catch (e) {
            this.constructor.prototype.logger.error(e);
            throw e;
        }
    }
    async findById(id) {
        try {
            const response = await this.constructor.prototype.client.get({
                id: this.id || id,
                index: this.constructor.prototype.__indexName,
                type: this.constructor.prototype.__type,
                realtime: true,
                refresh: true,
            });
            this.apply(response.body._source);
            return this;
        }
        catch (e) {
            this.constructor.prototype.logger.error(e);
            throw e;
        }
    }
    async save() {
        try {
            const response = await this.constructor.prototype.client.create({
                id: this.id || undefined,
                index: this.constructor.prototype.__indexName,
                type: this.constructor.prototype.__type,
                refresh: true,
                version_type: 'internal',
                body: this.toObject(),
            });
            this.apply(response.body);
            return this;
        }
        catch (e) {
            this.constructor.prototype.logger.error(e);
            throw e;
        }
    }
    async update() {
        try {
            const response = await this.constructor.prototype.client.update({
                id: this.id,
                index: this.constructor.prototype.__indexName,
                type: this.constructor.prototype.__type,
                refresh: true,
                body: { doc: this.toObject() },
            });
            this.apply(response.body);
            return this;
        }
        catch (e) {
            this.constructor.prototype.logger.error(e);
            throw e;
        }
    }
    async partialUpdate(data) {
        try {
            await this.constructor.prototype.client.update({
                id: this.id,
                index: this.constructor.prototype.__indexName,
                type: this.constructor.prototype.__type,
                refresh: true,
                body: { doc: Object.assign({}, data) },
            });
            return await this.get();
        }
        catch (e) {
            this.constructor.prototype.logger.error(e);
            throw e;
        }
    }
    async remove() {
        try {
            await this.constructor.prototype.client.delete({
                id: this.id,
                index: this.constructor.prototype.__indexName,
                type: this.constructor.prototype.__type,
                refresh: true,
            });
            return true;
        }
        catch (e) {
            throw e;
        }
    }
    async find(params, from, size, sort) {
        const query = this.generateQuery(params);
        let body = {
            index: this.constructor.prototype.__indexName,
            type: this.constructor.prototype.__type,
            body: {
                query: query,
            },
        };
        if (typeof from === 'number') {
            body = Object.assign(Object.assign({}, body), { from });
        }
        if (size) {
            body = Object.assign(Object.assign({}, body), { size });
        }
        if (sort) {
            body = Object.assign(Object.assign({}, body), { sort: this.generateSort(sort) });
        }
        try {
            const data = await this.constructor.prototype.client.search(Object.assign({}, body));
            const { body: { hits: { hits, total: { value }, }, }, } = data;
            const items = hits.map(({ _id, _source }) => {
                const doc = new this.constructor();
                doc.apply(Object.assign(Object.assign({}, _source), { id: _id }));
                return doc;
            });
            return { items, total: value };
        }
        catch (e) {
            this.constructor.prototype.logger.error(e);
            throw e;
        }
    }
    async removeWhereNotIn(term, array) {
        try {
            const body = {
                index: this.constructor.prototype.__indexName,
                type: this.constructor.prototype.__type,
                body: {
                    query: {
                        bool: {
                            must_not: {
                                terms: {
                                    [term]: array,
                                },
                            },
                        },
                    },
                },
            };
            const response = await this.constructor.prototype.client.delete_by_query(Object.assign({}, body));
            return response.body.deleted;
        }
        catch (e) {
            throw e;
        }
    }
    generateSort(sort) {
        return Object.keys(sort).map((f) => `${f}:${sort[f].toLowerCase()}`);
    }
    generateQuery(params) {
        if (typeof params === 'string') {
            return {
                multi_match: {
                    query: params,
                    fields: ['*'],
                },
            };
        }
        else {
            const filter = params;
            if (Object.keys(filter).length === 1) {
                return {
                    match: filter,
                };
            }
            const query = {
                bool: {
                    must: Object.keys(filter).map((k) => ({
                        match: { [k]: filter[k] },
                    })),
                },
            };
            return query;
        }
    }
    apply(data) {
        this.id = (data === null || data === void 0 ? void 0 : data.id) || this.id;
        Object.keys(data).forEach((attr) => {
            if (this.constructor.prototype.__fields.hasOwnProperty(attr)) {
                this[attr] = data[attr];
            }
        });
    }
    toObject() {
        try {
            let data = {};
            Object.keys(this.constructor.prototype.__fields).forEach((i) => {
                const field = this.constructor.prototype.__fields[i];
                data = Object.assign(Object.assign({}, data), { [field.fieldName]: this[field.fieldName] || field.default || undefined });
            });
            return data;
        }
        catch (e) {
            throw e;
        }
    }
    toJson() {
        try {
            return this.toObject();
        }
        catch (e) {
            throw e;
        }
    }
}
exports.BaseDocument = BaseDocument;
//# sourceMappingURL=base-index.model.js.map